import xchat

__module_author__ = "jvdm"
__module_name__ = "GetMonitor"
__module_version__ = "0.5"
__module_description__ = "List Server requests monitor for xchat."


###############################################################################
# Utilities Functions
###############################################################################

def same_dcc(d1, d2):
    return d1.port == d2.port and xchat.nickcmp(d1.nick, d2.nick) == 0 \
           and d1.file == d2.file


###############################################################################
# Utilities Classes
###############################################################################

class Output:
    """ Output create a text output interface for messages into a query window
    in xchat.

    """
    def __init__(self):
        self.query_name = ">>getmonitor<<"
        self.__getQuery()

    def __getQuery(self):
        """ Creates an output/input query for getmonitor. Returns the
        query context.

        This will open a new query on the current context when the
        script is loaded. In this query only getmonitor commands are
        accepted and all output will be send.

        """
        if not xchat.find_context(channel=self.query_name):
            xchat.command("QUERY %s" % self.query_name)
        return xchat.find_context(channel=self.query_name)

    def out(self, s):
        c = self.__getQuery()
        if xchat.nickcmp(
            xchat.get_context().get_info("channel"), 
            c.get_info("channel")) != 0:
            xchat.prnt(s)
        c.prnt(s)

o = Output() # The Output object we use during the script


###############################################################################
# The DCC Event Connector and Generator
###############################################################################

class DccEventGenerator:
    """ Use the xchat.timer to look for receiving dcc events

    The DccEventGenerator is an event driven timer that triggers an method to
    compare the dcc states with a internal mirror, creating a new xchat timer
    thread for each new function hooked to one of the events.

    new_event -- a new dcc has come
    changed_event -- the dcc has changed his status
    closed_event -- the dcc was closed
    """

    RECEIVING = 1

    def same_dcc(d1, d2):
        return d1.port == d2.port and xchat.nickcmp(d1.nick, d2.nick) == 0 \
               and d1.file == d2.file

    def __init__(self, loop_timer = 5000):
        self.mirror = [] 
        self.thook = xchat.hook_timer(loop_timer, self.__runner)
        self.handlers = {"new_event": {},
                         "changed_event": {}, 
                         "closed_event": {}}
        self.eid = {"new_event": 1, 
                    "changed_event": 2, 
                    "closed_event": 3}
    
    def connect(self, event, handler):
        if callable(handler):
            if self.handlers.has_key(event):
                handler_id = id(handler)
                # we generate the handler id by the object id
                # to guarantee uniqueness we check every handler
                while self.handlers[event].has_key(handler_id):
                    handler_id += 1
                self.handlers[event][handler_id] = handler
            else:
                raise ValueError, "unknown %s event." % event
            # that's the connect id used to unhook
            return handler_id
        else:
            raise TypeError("%s argument must be callable." % handler)

    def disconnect(self, handler_id):
        for hand_dic in self.handlers.values():
            if hand_dic.pop(handler_id, None):
                return True
        return False

    def changeTimer(self, nt):
        """ Change the verification timer to 'nt' """
        xchat.unhook(self.thook)
        self.thook = xchat.hook_timer(nt, self.__runner)
   
    def __runner(self, data=None):
        # new_mirror will contain the new dcc status
        new_mirror = []
        for xdcc in xchat.get_list("dcc"):
            # monitoring only receving dccs
            if xdcc.type == self.RECEIVING:
                # considering every dcc new until notice the contrary
                new = True
                for m in self.mirror:
                    # checking for new dcc
                    if same_dcc(xdcc, m):
                        new = False
                        if m.status != xdcc.status:
                            self.__emit("changed_event", xdcc)
                        # this will force only closed dccs to remain in
                        # self.mirror
                        self.mirror.remove(m)
                        break
                if new:
                    self.__emit("new_event", xdcc)
                # every new dcc will be here
                new_mirror.append(xdcc)                    
        for dcc in self.mirror:
            self.__emit("closed_event", dcc)
        self.mirror = new_mirror
        return True  # xchat.timer_hook need this to loop again
    
    def __emit(self, event, dcc):
        if self.handlers.has_key(event):
            for handler in self.handlers[event].values():
                if handler(dcc):
                    break
        else:
            raise ValueError, "invalid event '%s'" % event


DccEvent = DccEventGenerator() # the object we use during the script


###############################################################################
# Model Classes
###############################################################################

class Request:
    """ Represents a request made to some LsitServer/Channel"""
    
    def __init__(self, ls, fname, ext):
        self.ls = ls          # list server object
        self.fname =  fname   # the file's request filename
        self.ext = ext        # the file's extension
    
    def match(self, dcc):
        """ Match this request to a xchat dcc object """
        f = self.fname + self.ext
        if dcc.file == f or dcc.file == f.replace(" ", "_"):
            if xchat.nickcmp(self.ls.nick, dcc.nick) == 0:
                return True
        return False

    def delete(self):
        self.ls = None        

class Transfer(Request):
    """ Represents a request made to some LsitServer/Channel"""

    def __init__(self, request, dcc):
        Request.__init__(self, request.ls, request.fname, request.ext)
        self.dcc = dcc

    def match(self, dcc):
        return same_dcc(self.dcc, dcc)

    def __status_error(self, new):
        o.out("DEBUG: possible error -- status changed from "
        "%d to %d" % (self.dcc.status, new.status))
        o.out("DEBUG: server '%s': tranfer '%s'" % (self.ls.nick,
                                                    self.dcc.file))
    def statusChanged(self, new):
        if new.status == 2:
            o.out("DEBUG: dcc FAILED, was status 1, now status 2")
            o.out("DEBUG: lost the request.")
            self.ls.transfers.remove(self)
            return
        elif new.status == 5:
            o.out("DEBUG: '%s' from '%s' aborted." % (self.dcc.file,
                                                      self.dcc.nick))
            o.out("DEBUG: lost the request.")
            self.ls.transfers.remove(self)
            return

        if self.dcc.status == 0:   # queued
            if new.status in [1, 3, 4]:
                self.dcc = new              

        elif self.dcc.status == 1: # active
            if new.status == [0, 4]:
                self.__status_error(new)
            elif new.status == 3:
                self.dcc = new

        elif self.dcc.status == 3: # done
            self.__status_error(new)

        elif self.dcc.status == 4: # connecting
            if new.status == 0:
                self.__status_error(new)
            elif new.status in [1, 3]:
                pass

    def delete(self):
        Request.delete()
        self.dcc = None

class ListServer:
    def __init__(self, irc, channel, nick):
        self.irc = irc
        self.channel = channel
        self.nick = nick

        self.queue = []
        self.requests = []
        self.transfers = []

        self.flags = {"adopt_all_sends": False,
                      "queue_size": 1,
                      "extensions": [".mp3"],
                      "update_interval": 5000}

        self.new_conn = DccEvent.connect("new_event", self.newDccHandler)
        self.changed_conn = DccEvent.connect(
            "changed_event", self.changedDccHandler)

        self.update_timer = None

    def makeRequests(self, userdata):
        if not self.queue:
            xchat.unhook(self.update_timer)
            self.update_timer = None
        else:
            # TODO: make possible to stop timer if queue_size reached
            if len(self.requests) < self.flags["queue_size"]:
                filename = self.queue.pop(0)
                c = xchat.find_context(server=self.irc, channel=self.channel)
                # TODO: what if the channel is closed?
                if c:
                    c.command("say !%s %s" % (self.nick, filename))
                    ext = "." + filename.split(".")[-1]
                    filename = filename[:filename.find(ext)]
                    self.requests.append(Request(self, filename, ext))
                else:
                    o.out("Fuck, the server/channel is closed, nothing done.")
                    o.out("ListServer '%s', request '%s' to '%s@%s'" %\
                          (self.nick, filename, self.channel, self.irc))
                    o.out("Throwing out this request.")
        return True

    def delete(self):
        DccEvent.disconnect(self.new_conn)
        DccEvent.disconnect(self.changed_conn)
        self.queue = None
        for r in self.requests:
            r.delete()
            self.requests.remove(r)
        for t in self.transfers:
            t.delete()
            self.transfers.remove(t)

    def setFlag(self, flag, o):
        if self.flags.has_key(flag):
            self.flags[flag] = o;
        else:
            raise ValueError, "'%s' is not a valid ListServer flag"

    def addRequest(self, filename):
        if not self.queue:
            self.update_timer = xchat.hook_timer(self.flags["update_interval"],
                                                 self.makeRequests)
        self.queue.append(filename)

    def newDccHandler(self, dcc):
        if xchat.nickcmp(dcc.nick, self.nick) == 0:
            if self.flags["adopt_all_sends"]:
                o.out("<%s> adopting dcc send '%s'" % (self.nick, dcc.file))
                return True
            else:
                for f in self.requests:
                    if f.match(dcc):
                        o.out("<%s> got a new dcc '%s'" % (self.nick, 
                                                           dcc.file))
                        self.requests.remove(f)
                        self.transfers.append(Transfer(f, dcc))
                        # TODO: aceitar o dcc se nao foi aceito ainda
                        return True
        return False

    def changedDccHandler(self, dcc):
        if xchat.nickcmp(dcc.nick, self.nick) == 0:
            for t in self.transfers:
                if t.match(dcc):
                    t.statusChanged(dcc)
                    return True
        return False

    def closedDccHandler(self, dcc):
        if xchat.nickcmp(dcc.nick, self.nick) == 0:
            for t in self.transfers:
                if t.match(dcc):
                    self.transfers.remove(t)
                    return True
        return False
                    
class ListServerList:
    def __init__(self):
        self.irc = {}

    def __lowerArgs(self, irc, channel, nick):
        return (irc.lower(), channel.lower(), nick.lower())

    def findListServer(self, irc, channel, nick):
        irc, channel, nick = self.__lowerArgs(irc, channel, nick)
        try:
            for listserver in self.irc[irc][channel]:
                if xchat.nickcmp(listserver.nick, nick) == 0:
                    return listserver
        except KeyError:
            return None

    def addListServer(self, irc, channel, nick):
        ls = self.findListServer(irc, channel, nick)
        if not ls:
            irc, channel, nick = self.__lowerArgs(irc, channel, nick)
            ls = ListServer(irc, channel, nick)
            if not irc in self.irc.keys():
                self.irc[irc] = {}

            if not channel in self.irc[irc].keys():
                self.irc[irc][channel] = []
                
            self.irc[irc][channel].append(ls)
        return ls

    def removeListServer(self, irc, channel, nick):
        ls = self.findListServer(irc, channel, nick)
        if ls:
            ls.delete()
            self.irc[irc][channel].remove(ls)
            return True
        return False


###############################################################################
# Command Interface Class 
###############################################################################

class GetMonitorCommand:
    """ The command interface for GetMonitor. """

    def __init__(self):
        o.out("Starting GetMonitor %s" % __module_version__)
        xchat.hook_command("gmadd", self.gmaddCommand, "")
        xchat.hook_command("gmset", self.gmsetCommand, "")
        xchat.hook_command("gmlist", self.gmlistCommand, "")
        xchat.hook_command("gmpy", self.gmpyCommand, "")
        self.lslist = ListServerList()

    def gmpyCommand(self, word, word_eol, userdata):
        exec(word_eol[1])
        return xchat.EAT_ALL

    def gmlistCommand(self, word, word_eol, userdata):
        st = {0: "queued", 1: "active", 2: "failed", 3: "done", 
              4: "connecting", 5: "aborted"}
        got = False
        o.out("*** LISTING LIST SERVERS ***")
        for c in self.lslist.irc.values():
            for listserver in c.values():
                for ls in listserver:
                    o.out("<%s> -> %s@%s" % (ls.nick, ls.channel, ls.irc))
                    got = True
                    for t in ls.queue:
                        if got:
                            o.out("Not asked:")
                            got = False
                        o.out("    %s" % t)
                    got = True                    
                    for t in ls.requests:
                        if got:
                            o.out("Asked:")
                            got = False
                        o.out("    %s" % (t.fname + t.ext))
                    got = True
                    for t in ls.transfers:
                        if got:
                            o.out("Transfers:")
                            got = False
                        o.out("    [%s] %s" % (st[t.dcc.status], t.dcc.file))
                o.out("--------")
        o.out("****************************")
        return xchat.EAT_ALL

    def gmsetCommand(self, word, word_eol, userdata):
        if len(word) < 4:
            o.out("gmset error: bad command usage.")
        else:
            irc = xchat.get_info("host")
            channel = xchat.get_info("channel")
            ls = self.lslist.findListServer(irc, channel, word[1])
            if ls:
                if word[2] in ls.flags.keys():
                    try:
                        op = type(ls.flags[word[2]])(word[3])
                        ls.setFlag(word[2], op)
                        o.out("gmset: flag %s was set to %s" % (word[2], op))
                    except:
                        o.out("gmset error: bad options for flag %s" % word[2])
                else:
                    o.out("gmset error: '%s' is not a ListServer flag." % \
                          word[2])
            else:
                o.out("gmset error: %s not found in %s@%s" % (word[1], 
                                                              channel, irc))
        return xchat.EAT_ALL            

    def gmaddCommand(self, word, word_eol, userdata):
        """ Callback for the 'gmadd' command, adding a List Serve to be 
        monitored.

        By adding a List Server getmonitor can monitor requests and
        events by this server. All requests must be to added List
        Servers on a specified channel.

        """
        if len(word) < 3:
            o.out("gmadd error: bad command usage, try /help gmadd")
        else:
            channel = xchat.get_info("channel")
            irc = xchat.get_info("host")
            if irc:
                if channel[0] == "#":
                    nick = word[1][1:]
                    filename = word_eol[2]
                    ls = self.lslist.findListServer(irc, channel, nick)
                    if not ls:                           
                        ls = self.lslist.addListServer(irc, channel, nick)
                    ls.addRequest(filename)

                else:
                    o.out("%s isn't a channel." % channel)
            else:
                o.out("you must be connected to add requests")

        return xchat.EAT_ALL


GetMonitorCommand()
